# coding: utf-8

import os
import itertools
from datetime import datetime, timedelta
from functools import wraps
from pathlib import Path

from golgi.config import configurable
from tek.tools import first_valid, datetime_to_unix, find
from series.logging import Logging

from series import EpisodeMetadataFactory, is_episode, EpisodeMetadata

from series.library.model.collection import EpisodeCollection, MovieCollection
from series.library.model.series import Series
from series.library.model.season import Season
from series.library.model.episode import Episode
from series.library.model.watch_event import WatchEvent
from series.library.model.movie import Movie


@configurable(library=['collection_paths', 'movie_collection_paths'])
class LibraryFacade(Logging):

    @property
    def lock(self):
        return self._db.lock

    def commit(func):
        @wraps(func)
        def wrapper(self, *a, **kw):
            result = func(self, *a, **kw)
            self._db.commit()
            return result
        return wrapper

    def __init__(self, db_):
        self._db = db_
        self._epimeta_factory = EpisodeMetadataFactory()

    @property
    def _episode_collections(self):
        return [EpisodeCollection(path) for path in self._collection_paths]

    @property
    def _movie_collections(self):
        return [MovieCollection(path) for path in self._movie_collection_paths]

    def scan(self):
        self.scan_episodes()
        self.scan_movies()

    def scan_episodes(self):
        ''' Scan all present collections for new episodes.
        '''
        self.log.info('Rescanning episode collections…')
        files = itertools.chain(*[c.video_files for c in
                                  self._episode_collections])
        new_files = 0
        for _file in files:
            if is_episode(_file):
                data = self._epimeta_factory.from_filename(Path(_file))
                new = not self._episode_exists(data)
                if new:
                    new_files += 1
                self._add_new_episode(data, new=new)
        self.log.info('Done. Found {} new episodes.'.format(new_files))

    def scan_movies(self):
        self.log.info('Rescanning movie collections…')
        files = itertools.chain(*[c.video_files for c in
                                  self._movie_collections])
        new_files = 0
        for _file in files:
            title = os.path.splitext(os.path.basename(_file))[0]
            new_files += not self._movie_exists(title)
            self._add_new_movie(title)
        self.log.info('Done. Found {} new movies.'.format(new_files))

    @commit
    def clean(self):
        self.clean_episodes()
        self.clean_movies()

    def clean_episodes(self):
        ''' Iterate over all episodes in the db, marking those as
        removed that cannot be found in any collection.
        '''
        removed = 0
        for episode in self.episodes():
            if not any([c.contains_video(episode) for c in
                        self._episode_collections]):
                episode.removed = True
                removed += 1
        self.log.info('Done. Removed {} episodes.'.format(removed))

    def clean_movies(self):
        removed = 0
        for movie in self.movies():
            if not any([c.contains_video(movie) for c in
                        self._movie_collections]):
                movie.removed = True
                removed += 1
        self.log.info('Done. Removed {} movies.'.format(removed))

    def _episode_exists(self, data):
        return self.episode(series=data.series, season=data.season,
                            number=data.episode, removed=True) is not None

    def _movie_exists(self, title):
        return self.movie(title=title, removed=True) is not None

    @commit
    def _add_new_episode(self, data, new=True, tvdb_id=None):
        ''' Add an episode with corresponding series and season objects
        if they don't exist yet.
        If an episode was added that had been removed before, it is
        reactivated, but not marked as new.
        '''
        series = self._create_if_missing(Series, name=data.series, tvdb_id=tvdb_id)
        season = self._create_if_missing(Season, number=data.season, series=series)
        episode = self._create_if_missing(Episode, number=data.episode,
                                          series=series, season=season)
        episode.removed = False
        if new:
            episode.new = True
        return episode

    @commit
    def _add_new_movie(self, title):
        movie = self._create_if_missing(Movie, title=title)
        movie.removed = False
        return movie

    def _create_if_missing(self, Model, **params):
        ''' Call the DB factory function corresponding to _type with
        params, if there is no such object in the db yet.
        '''
        query = self._db.query(Model).filter_by(**params)
        return (query.first() if query.count() > 0 else self._create(Model,
                                                                     **params))

    def _create(self, Model, **params):
        model = Model(**params)
        self._db.add(model)
        return model

    def series(self, name):
        ''' Return the series with the specified name.
        '''
        return self._db.query(Series).filter_by(name=name).first()

    def episodes(self, series=None, season=None, numbers=None, removed=False,
                 extra={}):
        ''' Query the db for episodes, return all matches.
        If any of the arguments are specified, they are used to filter
        the results by series, season and episode number.
        '''
        query = self._db.query(Episode).filter_by(**extra)
        if not removed:
            query = query.filter_by(removed=False)
        if series is not None:
            if not isinstance(series, Series):
                series = self.series(series)
            query = query.filter_by(series=series)
        if season is not None:
            if not isinstance(season, Season):
                season = self.season(series, season)
            query = query.filter_by(season=season)
        if isinstance(numbers, list):
            query = query.filter(Episode.number.in_(numbers))
        return sorted(query.all())

    def episode(self, series, season, number, **kw):
        return first_valid(self.episodes(series=series, season=season,
                                         numbers=[number], **kw))

    def seasons(self, series=None, numbers=None, extra={}):
        query = self._db.query(Season).filter_by(**extra)
        if series is not None:
            if not isinstance(series, Series):
                series = self.series(series)
            query = query.filter_by(series=series)
        if isinstance(numbers, list):
            query = query.filter(Season.number.in_(numbers))
        return sorted(query.all())

    def season(self, series, number):
        return first_valid(self.seasons(series=series, numbers=[number]))

    def movies(self, title=None, removed=False, extra={}):
        query = self._db.query(Movie).filter_by(**extra)
        if not removed:
            query = query.filter_by(removed=False)
        if title is not None:
            query = query.filter_by(title=title)
        return sorted(query.all())

    def movie(self, title, **kw):
        return first_valid(self.movies(title=title, **kw))

    def episode_path(self, episode):
        results = [c.video_path(episode) for c in self._episode_collections]
        return first_valid(results)

    def episode_subtitle_path(self, episode):
        results = [c.subtitle_path(episode) for c in self._episode_collections]
        return first_valid(results)

    def movie_path(self, movie):
        results = [c.video_path(movie) for c in self._movie_collections]
        return first_valid(results)

    def movie_subtitle_path(self, movie):
        results = [c.subtitle_path(movie) for c in self._movie_collections]
        return first_valid(results)

    def video_path(self, item):
        if isinstance(item, Episode):
            return self.episode_path(item)
        elif isinstance(item, Movie):
            return self.movie_path(item)

    def subtitle_path(self, item):
        if isinstance(item, Episode):
            return self.episode_subtitle_path(item)
        elif isinstance(item, Movie):
            return self.movie_subtitle_path(item)

    @property
    def new_episodes(self):
        return sorted(filter(lambda e: e.new, self.episodes()))

    @property
    def all_series(self):
        return [s for s in self._db.query(Series).all() if not s.empty]

    @property
    def all_movies(self):
        return self.movies()

    def add_watch_event(self, series, season, number, start_time, finish_time,
                        stopped_at):
        episode = self.episode(series, season, number)
        event = WatchEvent(time_begin=datetime_to_unix(start_time),
                           time_end=datetime_to_unix(finish_time),
                           episode=episode, stopped_at=stopped_at)
        episode.new = False
        self._db.session.add_then_commit(event)
        return event

    def recently_watched_episodes(self, days):
        thresh = datetime_to_unix(datetime.now() - timedelta(days=days))
        events = self._db.query(WatchEvent).filter(WatchEvent.time_begin >
                                                   thresh)
        episodes = [event.episode for event in events]
        return sorted(set(episodes), reverse=True, key=lambda e:
                      e.last_watched)

    def recently_watched_seasons(self, show, season, days):
        thresh = datetime_to_unix(datetime.now() - timedelta(days=days))
        events = (
            self._db.query(WatchEvent)
            .filter(WatchEvent.time_begin > thresh)
            .join(Episode, WatchEvent.episode_id == Episode.id)
            .join(Season, Episode.season_id == Season.id)
            .join(Series, Episode.series_id == Series.id)
            .filter(Series.name == show)
            .filter(Season.number == int(season))
        )
        return [event.episode for event in events]

    def next_episode(self, episode):
        comp = lambda l, r: l > r
        return self._cycle(episode, comp, sorted)

    def previous_episode(self, episode):
        comp = lambda l, r: l < r
        reverse = lambda seq: sorted(seq, reverse=True)
        return self._cycle(episode, comp, reverse)

    def _cycle(self, episode, comp, sorter):
        season = episode.season
        candidate = find(lambda e: comp(e, episode), sorter(season.episodes))
        if candidate is None:
            series = episode.series
            next_season = find(lambda s: comp(s, season),
                               sorter(series.seasons))
            if next_season is not None and next_season.episodes.count() > 0:
                candidate = sorter(next_season.episodes)[0]
        return candidate

    def create_episode(self, series, season, episode, tvdb_id=None):
        data = EpisodeMetadata(series, season, episode)
        return self._add_new_episode(data, tvdb_id=tvdb_id)

    @commit
    def alter_episode(self, series, season, episode, data):
        episode = self.episode(series, season, episode, removed=True)
        if episode:
            for key, value in data.items():
                setattr(episode, key, value)
        return episode

    @commit
    def delete_episode(self, series, season, episode, data):
        episode = self.episode(series, season, episode, removed=True)
        if episode:
            self._db.delete(episode)
            return True

    @commit
    def alter_season(self, series, season, data):
        season = self.season(series, season)
        if season:
            for key, value in data.items():
                setattr(season, key, value)
        return season

    @commit
    def alter_series(self, series, data):
        series = self.series(series)
        if series:
            for key, value in data.items():
                setattr(series, key, value)
        return series

    def alter_object(self, item, data):
        if isinstance(item, Episode):
            self.alter_episode(item.series, item.season, item.number, data)
        elif isinstance(item, Season):
            self.alter_season(item.series, item.number, data)

__all__ = ['LibraryFacade']



"""add downgrade_after

Revision ID: 18b0a0874c5a
Revises: d363aeb3c4de
Create Date: 2016-09-28 03:00:45.197012

"""

# revision identifiers, used by Alembic.
revision = '18b0a0874c5a'
down_revision = 'd363aeb3c4de'

from alembic import op
import sqlalchemy as sa


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.add_column('release_monitor', sa.Column('downgrade_after', sa.Integer(), nullable=True))
    op.add_column('release_monitor', sa.Column('last_torrent_update_stamp', sa.Integer(), nullable=True))
    op.add_column('show', sa.Column('downgrade_after', sa.Integer(), nullable=True))
    op.add_column('torrent', sa.Column('caching', sa.Boolean(), nullable=True))
    ### end Alembic commands ###


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.add_column('release_monitor', sa.Column('torrent_cached', sa.BOOLEAN(), nullable=True))
    op.add_column('link', sa.Column('real_type', sa.VARCHAR(length=24), nullable=False))
    ### end Alembic commands ###



from series.handler import Handler

class BaseHandler(Handler):

    def __init__(self, data, interval, description, **kw):
        self._data = data
        super().__init__(interval, description, **kw)

    @property
    def _lock(self):
        return self._data.lock

__all__ = ['BaseHandler']



from typing import Callable

from amino import _, Map, List, L
from amino.util.string import camelcaseify

from golgi.io.terminal import terminal as term, ColorString


def explain_item(e):
    for comp in e:
        name = str(ColorString(comp['name'], term.blue))
        cond = comp['cond']
        yield '{}: {}'.format(name, cond)


def explain_lines(data, title: Callable):
    for r in data:
        item = r['item']
        yield str(ColorString(title(item), term.bold))
        yield from explain_item(r['expl'])
        yield ''


def explain_release_lines(data):
    def title(monitor):
        data = Map(monitor['release']).values_at('series', 'season', 'episode')
        return '{} {}x{}'.format(*data)
    return List.wrap(explain_lines(data, title))


def format_explain_release(data):
    return explain_release_lines(data).join_lines


def explain_show_lines(data):
    return List.wrap(explain_lines(data, _['name']))


def format_explain_show(data):
    return explain_show_lines(data).join_lines


def format_status_lines(output):
        ind = '  {}'.format
        id = lambda r: ColorString('#{}'.format(r['id']), term.red)
        series = lambda r: ColorString(
            camelcaseify(r['series'], sep=' ', splitter='[ _]'), term.yellow)
        head = lambda t: ColorString('{}:'.format(t), term.green)
        def enum(r):
            return '{}x{}'.format(
                ColorString(r['season'], term.blue),
                ColorString(r['episode'], term.blue),
            )
        desc = lambda r: '{} {} {}'.format(id(r), series(r), enum(r))
        date = lambda r, d: ColorString(r.get(d, 'no date'), term.bold)
        with_date = lambda r, d: ind('{} – {}'.format(date(r, d), desc(r)))
        ad = lambda r: with_date(r, 'airdate')
        dd = lambda r: with_date(r, 'download_date')
        any_date = lambda r: r.get('download_date', r.get('airdate', ''))
        def entry(rs, h, fmt):
            s = List.wrap(rs).sort_by(any_date)
            return (s / fmt).cons(head(h)) if rs else List()
        def section(key, fmt, head):
            return output.lift(key) / L(entry)(_, fmt, head) | List()
        return List(
            ('done', 'Downloaded episodes', dd),
            ('caching', 'Caching releases', ad),
            ('search', 'Searching for torrents', ad),
            ('next', 'Upcoming episodes', ad),
            ('downloading', 'Currently downloading', ad),
            ('failed', 'Failed downloads', ad),
            ('current_search', 'Current torrent search', ad),
        ).flat_map3(section)


def format_status(output):
    return format_status_lines(output).join_lines

__all__ = ('explain_item', 'explain_lines', 'explain_release_lines',
           'format_explain_release', 'explain_show_lines',
           'format_explain_show')



import abc
import requests
import functools
from typing import Callable, Any

from flask import json

from amino import IO, Left, Right, List, __, _, L, Empty, Just, Boolean

from series.logging import Logging

from series.client.errors import SeriesClientException


def is_int(value):
    return (isinstance(value, int) or isinstance(value, str) and
            value.isnumeric())


class ApiClient(Logging):

    def __init__(self, url, port, info_output=True):
        self.url = url
        self.port = port
        self._info_output = info_output

    def _url(self, path):
        return '{}:{}/{}'.format(self.url, self.port, path)

    def _request(self, req_type, path, body):
        headers = {'content-type': 'application/json'}
        requester = getattr(requests, req_type)
        url = self._url(path)
        self.log.debug('api request: {}'.format(url))
        try:
            response = requester(url, data=json.dumps(body), headers=headers)
        except requests.RequestException as e:
            msg = 'Request failed! ({})'.format(e)
            raise SeriesClientException(msg) from e
        else:
            if response.status_code >= 400:
                self.log.error(
                    'API response status {}'.format(response.status_code))
            try:
                _json = response.json()
            except ValueError as e:
                msg = 'Error in API request (no JSON in response)!'
                raise SeriesClientException(msg) from e
            else:
                data = _json.get('response', {})
                if isinstance(data, dict) and 'error' in data:
                    self.log.error(data['error'])
                return data

    def get(self, path, body={}):
        return self._request('get', path, body)

    def post(self, path, body={}):
        return self._request('post', path, body)

    def put(self, path, body={}):
        return self._request('put', path, body)

    def delete(self, path, body={}):
        return self._request('delete', path, body)


class Command(metaclass=abc.ABCMeta):

    def __init__(self, name, args, client, param_desc, desc) -> None:
        self.name = name
        self.args = List.wrap(args)
        self.client = client
        self.param_desc = param_desc
        self.desc = desc


class RecordCommand(Command):

    def __init__(self, name, args, client, param_desc, desc, params) -> None:
        super().__init__(name, args, client, param_desc, desc)
        self.params = params

    @abc.abstractproperty
    def _id_spec(self) -> IO[Left]:
        ...

    @abc.abstractproperty
    def _data_spec(self) -> IO[Right]:
        ...

    @abc.abstractproperty
    def _data_path(self) -> IO[str]:
        ...

    @abc.abstractproperty
    def _type(self) -> str:
        ...

    @abc.abstractproperty
    def rest(self):
        ...

    def _lookup(self, data):
        def check(resp):
            return (
                IO.failed('no release found')
                if resp == -1 else
                IO.pure((resp, self.rest))
            )
        return IO.delay(self.client.get, '{}_id'.format(self._type), body=data) // check

    @property
    def is_id_spec(self) -> Boolean:
        return self.args.head.exists(is_int)

    @property
    def spec(self):
        return (self._id_spec
                if self.is_id_spec else
                self._data_spec)

    @property
    def data(self):
        return (
            self.spec //
            __.lmap(lambda a: IO.pure((a, self.args.tail | List())))
            .left_or_map(self._lookup)
        )

    def run(self, f: Callable[[int, List], Any]) -> IO:
        return self.data.flat_map2(f)

    def _req(self, meth, path=Empty(), data=None):
        sub = path / '/{}'.format | ''
        cb = (lambda a: Just(dict())) if data is None else data
        def io(i, body):
            return IO.delay(meth(self.client), '{}/{}{}'.format(self._type, i, sub), body=body)
        def assemble(i, a):
            return cb(a).io('invalid arguments') // L(io)(i, _)
        def check_error(response):
            if isinstance(response, dict) and 'error' in response:
                return IO.failed(response['error'])
            else:
                return IO.pure(response)
        return self.run(assemble) // check_error

    def get(self, data: Callable=None, path=Empty()):
        return self._req(_.get, path=path, data=data)

    def put(self, data: Callable=None, path=Empty()):
        return self._req(_.put, path=path, data=data)

    def post(self, data: Callable=None, path=Empty()):
        return self._req(_.post, path=path, data=data)

    @property
    def delete(self):
        return self._req(_.delete)

    def create(self, data: Callable=None, path=Empty()):
        cb = (lambda a: Just(dict())) if data is None else data
        sub = path / '/{}'.format | ''
        def assemble(path):
            return (cb(self.rest).io('invalid arguments') /
                    L(self.client.post)('{}{}'.format(path, sub), _))
        return self._data_path // assemble


def command_base(Cmd, param_desc, desc, **extra):
    def decor(func):
        func._doc = (param_desc, desc)
        @functools.wraps(func)
        def wrap(self, *args):
            cmd = Cmd(func.__name__, args, self.client, param_desc, desc,
                      **extra)
            return func(self, cmd).attempt
        return wrap
    return decor


def command(param_desc, desc):
    return command_base(Command, param_desc, desc)


class ApiClientMeta(abc.ABCMeta):

    def __new__(cls, name, bases, namespace) -> None:
        inst = super().__new__(cls, name, bases, namespace)  # type: ignore
        doc = ((name, fun._doc) for name, fun in namespace.items()
               if hasattr(fun, '_doc'))
        inst.doc = dict(doc)
        return inst


class ClientBase(Logging):

    @abc.abstractproperty
    def client(self):
        ...

    @command('', 'Display this help text')
    def help(self, cmd):
        if self.doc:
            maxlen = len(max(self.doc.keys(), key=len))
            pad = lambda s: s.ljust(maxlen)
            def output():
                yield 'Available commands:'
                data = List.wrap(self.doc.items()).sort_by(_[0])
                for name, (param_desc, desc) in data:
                    yield ''
                    yield '{}    {}'.format(pad(name), param_desc)
                    yield '  {}'.format(desc)
            return IO.delay(output) / '\n'.join
        else:
            return IO.failed('no doc for client')

__all__ = ['ApiClient']



from datetime import datetime

from amino import Dat, List, Maybe, Either, Try
from amino.json.data import JsonError

from series.util import datetime_to_unix


class ShowCodec(Dat['ShowCodec']):

    def __init__(self, id: int, seriesName: str, status: str) -> None:
        self.id = id
        self.seriesName = seriesName
        self.status = status


class SummaryCodec(Dat['SummaryCodec']):

    def __init__(self, airedEpisodes: int, airedSeasons: List[int]) -> None:
        self.airedEpisodes = airedEpisodes
        self.airedSeasons = airedSeasons


class EpisodeCodec(Dat['EpisodeCodec']):

    def __init__(self, airedEpisodeNumber: int, airedSeason: int, firstAired: str) -> None:
        self.airedEpisodeNumber = airedEpisodeNumber
        self.airedSeason = airedSeason
        self.firstAired = firstAired


class EpisodeMetadataCodec(Dat['EpisodeMetadataCodec']):

    def __init__(self, airedEpisodeNumber: int, airedSeason: int, episodeName: str, overview: str) -> None:
        self.airedEpisodeNumber = airedEpisodeNumber
        self.airedSeason = airedSeason
        self.episodeName = episodeName
        self.overview = overview


class TvdbShow(Dat['TvdbShow']):

    @staticmethod
    def from_codec(data: ShowCodec) -> 'TvdbShow':
        return TvdbShow(data.id, data.seriesName, data.status.lower() == 'ended')

    def __init__(self, id: int, name: str, ended: bool) -> None:
        self.id = id
        self.name = name
        self.ended = ended


class TvdbEpisode(Dat['TvdbEpisode']):

    @staticmethod
    def from_codec(data: EpisodeCodec) -> 'TvdbEpisode':
        airdate = Try(lambda: datetime.fromisoformat(data.firstAired))
        stamp = airdate.map(datetime_to_unix)
        return TvdbEpisode(data.airedEpisodeNumber, data.airedSeason, stamp.to_maybe)

    def __init__(self, number: int, season: int, airdate: Maybe[int]) -> None:
        self.number = number
        self.season = season
        self.airdate = airdate


class TvdbSeason(Dat['TvdbSeason']):

    def __init__(self, number: int, episodes: List[TvdbEpisode]) -> None:
        self.number = number
        self.episodes = episodes


class TvdbSeasonSummary(Dat['TvdbSeasonSummary']):

    def __init__(self, episodes: int, seasons: int) -> None:
        self.episodes = episodes
        self.seasons = seasons


class TvdbEpisodeMetadata(Dat['TvdbEpisodeMetadata']):

    @staticmethod
    def from_codec(data: EpisodeMetadataCodec) -> 'TvdbEpisodeMetadata':
        return TvdbEpisodeMetadata(data.airedEpisodeNumber, data.airedSeason, data.episodeName, data.overview)

    def __init__(self, number: int, season: int, title: str, overview: str) -> None:
        self.number = number
        self.season = season
        self.title = title
        self.overview = overview


__all__ = ('TvdbShow', 'TvdbEpisode', 'TvdbSeason', 'TvdbSeasonSummary', 'ShowCodec', 'SummaryCodec', 'EpisodeCodec',
           'EpisodeMetadataCodec', 'TvdbEpisodeMetadata',)
